import contextlib
import logging
import queue
import threading
import typing

T = typing.TypeVar("T")


class WorkerRunner:
    """
    Runner

    :param int parallelism: Number to of workers to run in parallel
    :param handler: Handler that receives work and optionally yields other work
    :param resource: Resource factory. Can be used to provide thread-specific resources.

    Example:
        def work(x: int):
            print(x)
            while 0 < x:
                x -= 1
                yield x

        runner = WorkerRunner(2, work)
        runner.run([3])

        # 3
        # 2
        # 1
        # 0
        # 0
        # 1
        # 0
        # 0
    """

    def __init__(self, parallelism: int, handler, resource=contextlib.nullcontext):
        self._handler = handler
        self._parallelism = parallelism
        self._resource = resource

    def run(self, items):
        queue_ = queue.LifoQueue()
        status = WorkStatus()

        for item in reversed(items):
            queue_.put(WorkItem(item))

        threads = []
        workers = []
        for _ in range(self._parallelism):
            worker = Worker(queue_, status, self._handler, self._resource)
            workers.append(worker)
            thread = threading.Thread(target=worker.run)
            thread.start()
            threads.append(thread)
        queue_.join()
        for _ in range(self._parallelism):
            queue_.put(END_ITEM)
        for thread in threads:
            thread.join()
        for worker in workers:
            worker.result()


"""
End item
"""
END_ITEM = None


class WorkItem(typing.Generic[T]):
    def __init__(self, value: T):
        self.value = value


class WorkStatus:
    def __init__(self):
        self.running = True


class Worker(typing.Generic[T]):
    def __init__(self, queue: queue.Queue, status: WorkStatus, handler, resource):
        super().__init__()

        self._queue = queue
        self._exception = None
        self._handler = handler
        self._resource = resource
        self._status = status

    def _drain(self):
        """
        Remove items from queue
        """
        while True:
            try:
                if self._queue.get(False) is END_ITEM:
                    break
            except queue.Empty:
                break
            else:
                self._queue.task_done()

    def run(self):
        try:
            with self._resource() as resource:
                self._work(resource)
        except Exception as e:
            self._exception = e
            self._status.running = False
        self._drain()

    def result(self):
        if self._exception is not None:
            raise self._exception

    def _work(self, resource):
        while self._status.running:
            item = self._queue.get()
            try:
                if item is END_ITEM:
                    self._shutdown = True
                    return
                result = self._handler(item.value, resource)
                if result is not None:
                    for item in result:
                        self._queue.put(WorkItem(item))
            finally:
                self._queue.task_done()



import hashlib
import importlib.resources as pkg_resources
import typing

from .collection.dict import groups

T = typing.TypeVar("T")


class Choice:
    def __init__(self, options: typing.List[T]):
        self._options = options

    def choose(self, input: bytes):
        b = hashlib.md5(input).digest()
        i = int.from_bytes(b[0:8], "big") % len(self._options)
        return self._options[i]


class GivenNameTransform:
    def __init__(self):
        with pkg_resources.open_text("slice_db.data", "given-name.txt") as f:
            text = f.read()
        options = [name for name in text.split("\n") if name]
        self._choice = Choice(options)

    def transform(self, text: str, pepper: bytes):
        return self._choice.choose(text.encode("utf-8") + pepper)


class SurnameTransform:
    def __init__(self):
        with pkg_resources.open_text("slice_db.data", "surname.txt") as f:
            text = f.read()
        options = [name for name in text.split("\n") if name]
        self._choice = Choice(options)

    def transform(self, text: str, pepper: bytes):
        return self._choice.choose(text.encode("utf-8") + pepper)


class GeozipTransform:
    def __init__(self):
        with pkg_resources.open_text("slice_db.data", "zip.txt") as f:
            text = f.read()
        options = [int(line) for line in text.split("\n") if line]
        g = groups(options, lambda x: str(x).zfill(5)[0:3])
        self._choices = {geozip: Choice(options) for geozip, options in g.items()}
        self._all_choices = Choice(options)

    def transform(self, zip: str, pepper: bytes):
        geo = zip[0:3]
        if geo not in self._choices:
            result = self._all_choices.choose(zip.encode("utf-8") + pepper)
        else:
            result = self._choices[geo].choose(zip.encode("utf-8") + pepper)
        return str(result).zfill(5)


def create_transform(type):
    if type == "geozip":
        return GeozipTransform()
    if type == "given_name":
        return GivenNameTransform()
    if type == "surname":
        return SurnameTransform()
    raise Exception(f"Invalid transform type {type}")



import contextlib
import threading
import typing

from ..graph import DepFn, check_cycle
from .work import WorkerRunner

T = typing.TypeVar("T", bound=typing.Hashable)


class ActionNode(typing.Generic[T]):
    def __init__(self, value: T, dependency_count: int):
        self.value = value
        self.reverse_dependencies = []
        self._dependency_count = dependency_count
        self._lock = threading.Lock()

    def can_execute(self):
        return not self._dependency_count

    def remove_dep(self):
        with self._lock:
            self._dependency_count -= 1
            return self.can_execute()


class GraphRunner:
    def __init__(self, parallelism, handler, resource=contextlib.nullcontext):
        self._runner = WorkerRunner(parallelism, self._handle_node, resource)
        self._handler = handler

    def run(self, items: typing.List[T], dep_fn: DepFn[T]):
        check_cycle(items, dep_fn)

        nodes = {item: ActionNode(item, len(dep_fn(item))) for item in items}

        for item, node in nodes.items():
            for dep in dep_fn(item):
                nodes[dep].reverse_dependencies.append(node)

        self._runner.run([node for node in nodes.values() if node.can_execute()])

    def _handle_node(self, item: ActionNode, resource):
        self._handler(item.value, resource)
        for dep in item.reverse_dependencies:
            if dep.remove_dep():
                yield dep



"""
Slice definitions
"""

import codecs
import typing
import zipfile

_MANIFEST_PATH = "manifest.json"


_UTF8_READER: codecs.StreamWriter = codecs.getreader("utf-8")


_UTF8_WRITER: codecs.StreamWriter = codecs.getwriter("utf-8")


def _segment_path(table_id: str, index: int) -> str:
    return f"{table_id}/{index + 1}.tsv"


class SliceReader:
    """
    Read slice
    """

    def __init__(self, file: typing.BinaryIO):
        self._zip = zipfile.ZipFile(file)

    def __enter__(self, *args, **kwargs):
        self._zip.__enter__(*args, **kwargs)
        return self

    def __exit__(self, *args, **kwargs):
        self._zip.__exit__(*args, **kwargs)

    def open_manifest(self) -> typing.ContextManager[typing.TextIO]:
        """
        Open manifest
        """
        file = self._zip.open(_MANIFEST_PATH)
        return _UTF8_READER(file)

    def open_schema(self, section: str):
        return self._zip.open(f"{section}.sql")

    def open_segment(
        self, table_id: str, index: int
    ) -> typing.ContextManager[typing.BinaryIO]:
        """
        Open segment
        """
        return self._zip.open(_segment_path(table_id, index))


class SliceWriter:
    """
    Write slice
    """

    def __init__(self, file: typing.BinaryIO):
        self._zip = zipfile.ZipFile(file, "w", compression=zipfile.ZIP_DEFLATED)

    def __enter__(self, *args, **kwargs):
        self._zip.__enter__(*args, **kwargs)
        return self

    def __exit__(self, *args, **kwargs):
        self._zip.__exit__(*args, **kwargs)

    def open_manifest(self) -> typing.ContextManager[typing.TextIO]:
        """
        Open manifest
        """
        file = self._zip.open(_MANIFEST_PATH, "w")
        return _UTF8_WRITER(file)

    def open_schema(self, section: str):
        return self._zip.open(f"{section}.sql", "w")

    def open_segment(
        self, table_id: str, index: int
    ) -> typing.ContextManager[typing.BinaryIO]:
        """
        Open segment
        """
        return self._zip.open(_segment_path(table_id, index), "w", force_zip64=True)



import dataclasses

import dataclasses_json

from ..json import DataJsonFormat, package_json_format


@dataclasses_json.dataclass_json(letter_case=dataclasses_json.LetterCase.CAMEL)
@dataclasses.dataclass
class Transform:
    pass


TRANSFORM_JSON_FORMAT = package_json_format("slice_db.formats", "manifest.json")


TRANSFORM_DATA_JSON_FORMAT = DataJsonFormat(TRANSFORM_JSON_FORMAT, Transform.schema())



import json
import os
import subprocess

from file import temp_file
from pg import connection, transaction
from process import run_process

_SCHEMA_SQL = """
    CREATE TABLE parent (
        id int PRIMARY KEY
    );

    CREATE TABLE child (
        id int PRIMARY KEY,
        parent_id int REFERENCES parent (id)
    );
"""

_SCHEMA_JSON = {
    "references": [
        {
            "columns": ["parent_id"],
            "id": "public.child.child_parent_id_fkey",
            "referenceColumns": ["id"],
            "referenceTable": "public.parent",
            "table": "public.child",
        }
    ],
    "tables": [
        {
            "columns": ["id"],
            "id": "public.parent",
            "name": "parent",
            "schema": "public",
        },
        {
            "columns": ["id", "parent_id"],
            "id": "public.child",
            "name": "child",
            "schema": "public",
        },
    ],
}

def test_dump(pg_database, snapshot):
    with temp_file("schema-") as schema_file, temp_file("output-") as output_file:
        with connection("") as conn, transaction(conn) as cur:
            cur.execute(_SCHEMA_SQL)

            cur.execute(
                """
                    INSERT INTO parent (id)
                    VALUES (1), (2);

                    INSERT INTO child (id, parent_id)
                    VALUES (1, 1), (2, 1), (3, 2);
                """
            )

        with open(schema_file, "w") as f:
            json.dump(_SCHEMA_JSON, f)

        run_process(
            [
                "slicedb",
                "dump",
                "--schema",
                schema_file,
                "--root",
                "public.parent",
                "id = 1",
                "--output",
                output_file,
            ]
        )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute(
                """
                    DELETE FROM child;

                    DELETE FROM parent;
                """
            )

        run_process(
            [
                "slicedb",
                "restore",
                "--input",
                output_file,
            ]
        )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("TABLE parent")
            result = cur.fetchall()
            assert result == [(1,)]

            cur.execute("TABLE child")
            result = cur.fetchall()
            assert result == [(1, 1), (2, 1)]

def test_dump_schema(pg_database, snapshot):
    with temp_file("schema-") as schema_file, temp_file("output-") as output_file:
        with connection("") as conn, transaction(conn) as cur:
            cur.execute(_SCHEMA_SQL)

            cur.execute(
                """
                    INSERT INTO parent (id)
                    VALUES (1), (2);

                    INSERT INTO child (id, parent_id)
                    VALUES (1, 1), (2, 1), (3, 2);
                """
            )

        with open(schema_file, "w") as f:
            json.dump(_SCHEMA_JSON, f)

        run_process(
            [
                "slicedb",
                "dump",
                "--include-schema",
                "--schema",
                schema_file,
                "--root",
                "public.parent",
                "id = 1",
                "--output",
                output_file,
            ]
        )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute(
                """
                    DROP TABLE child;

                    DROP TABLE parent;
                """
            )

        run_process(
            [
                "slicedb",
                "restore",
                "--include-schema",
                "--input",
                output_file,
            ]
        )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("TABLE parent")
            result = cur.fetchall()
            assert result == [(1,)]

            cur.execute("TABLE child")
            result = cur.fetchall()
            assert result == [(1, 1), (2, 1)]



import json

from ..dump import Schema, Table
from ..formats.dump import DumpReferenceDirection, DumpSchema
from .common import open_str_read, open_str_write


def filter_main(args):
    with open_str_read(args.input) as f:
        schema_json = json.load(f)
    input_config = DumpSchema.schema().load(schema_json)

    if args.subcommand == "children":
        output_config = children(args, input_config)

    output_json = DumpSchema.schema().dump(output_config)
    with open_str_write(args.output) as f:
        json.dump(output_json, f, sort_keys=True, indent=2)


def children(args, input_config: DumpSchema):
    schema = Schema(input_config)

    references = {reference.id: reference for reference in input_config.references}

    child_ids = set()

    def visit(table: Table):
        if table.id in child_ids:
            return
        child_ids.add(table.id)
        for reference in table.reverse_references:
            if DumpReferenceDirection.REVERSE in reference.directions:
                visit(reference.table)

    for table_id in args.table:
        visit(schema.get_table(table_id))

    for table in schema._tables.values():
        if table.id in child_ids:
            continue
        for reference in table.reverse_references:
            if reference.table.id in child_ids:
                references[reference.id].directions.remove(
                    DumpReferenceDirection.REVERSE
                )

    return input_config



import json
import os
import subprocess

from pg import connection, transaction
from process import run_process


def test_schema(pg_database, snapshot):
    with connection("") as conn, transaction(conn) as cur:
        cur.execute(
            """
                    CREATE TABLE parent (
                        id int PRIMARY KEY
                    );

                    CREATE TABLE child (
                        id int PRIMARY KEY,
                        parent_id int REFERENCES parent (id)
                    );
                """
        )

    result = run_process(["slicedb", "schema"])
    schema_json = json.loads(result)
    snapshot.assert_match(schema_json)



import codecs
import contextlib
import typing

import psycopg2.sql as sql

_UTF_8_WRITER = codecs.getwriter("utf-8")


class SqlWriter:
    def __init__(self, context, file):
        self._context = context
        self._file = file

    def open_predata(self):
        return contextlib.nullcontext(self._file)

    def open_postdata(self):
        return contextlib.nullcontext(self._file)

    @contextlib.contextmanager
    def open_data(
        self, id: str, index: int, schema: str, table: str, columns: typing.List[str]
    ):
        text_writer = _UTF_8_WRITER(self._file)

        text_writer.write(f"--\n-- Data for {id}/{index}\n--\n")

        query = sql.SQL("COPY {} ({}) FROM stdin;\n").format(
            sql.Identifier(schema, table),
            sql.SQL(", ").join(sql.Identifier(column) for column in columns),
        )
        text_writer.write(query.as_string(self._context))

        yield self._file

        text_writer.write("\\.\n\n\n")



import subprocess

from process import run_process


def test_transform_field_geozip():
    result = run_process(
        [
            "slicedb",
            "transform-field",
            "--pepper",
            "abc",
            "--transform",
            "geozip",
            "94356",
        ]
    )
    assert result.decode("utf-8") == "94303\n"


def test_transform_field_given_name():
    result = run_process(
        [
            "slicedb",
            "transform-field",
            "--pepper",
            "abc",
            "--transform",
            "given_name",
            "Jane",
        ]
    )
    assert result.decode("utf-8") == "Jeramy\n"


def test_transform_field_surname():
    result = run_process(
        [
            "slicedb",
            "transform-field",
            "--pepper",
            "abc",
            "--transform",
            "surname",
            "Baker",
        ]
    )
    assert result.decode("utf-8") == "Bertrand\n"



import sys


def open_bytes_read(path):
    return open(path, "rb") if path != "-" else sys.stdin.buffer


def open_bytes_write(path):
    return open(path, "wb") if path != "-" else sys.stdout.buffer


def open_str_read(path):
    return open(path, "r") if path != "-" else sys.stdin


def open_str_write(path):
    return open(path, "w") if path != "-" else sys.stdout



def query_schema(cur):
    """
    Query PostgreSQL for schema
    """
    cur.execute(
        """
            SELECT
                json_build_object(
                    'references', pc2.references,
                    'tables', pc.tables
                )
            FROM
                (
                    SELECT
                        coalesce(
                            json_agg(
                                json_build_object(
                                    'columns', pa.columns,
                                    'id', pn.nspname || '.' || pc.relname,
                                    'name', pc.relname,
                                    'schema', pn.nspname
                                )
                            ),
                            json '[]'
                        ) AS tables
                    FROM
                        pg_class AS pc
                        JOIN pg_namespace AS pn ON pc.relnamespace = pn.oid
                        CROSS JOIN LATERAL (
                            SELECT json_agg(pa.attname ORDER BY pa.attnum) AS columns
                            FROM pg_attribute AS pa
                            WHERE pc.oid = pa.attrelid AND 0 < pa.attnum AND NOT pa.attisdropped
                        ) AS pa
                    WHERE
                        pn.nspname <> 'information_schema'
                        AND pn.nspname NOT LIKE 'pg_%'
                        AND pc.relkind = 'r'
                ) AS pc
                CROSS JOIN (
                    SELECT
                        coalesce(
                            json_agg(
                                json_build_object(
                                    'columns', pa.columns,
                                    'id', pn.nspname || '.' || pc2.relname || '.' || pc.conname,
                                    'referenceColumns', pa.reference_columns,
                                    'referenceTable', pn2.nspname || '.' || pc3.relname,
                                    'table', pn.nspname || '.' || pc2.relname
                                )
                            ),
                            json '[]'
                        ) AS references
                    FROM
                        pg_constraint AS pc
                        JOIN pg_class AS pc2 ON pc.conrelid = pc2.oid
                        JOIN pg_namespace AS pn ON pc2.relnamespace = pn.oid
                        JOIN pg_class AS pc3 ON pc.confrelid = pc3.oid
                        JOIN pg_namespace AS pn2 ON pc3.relnamespace = pn2.oid
                        CROSS JOIN LATERAL (
                            SELECT
                                json_agg(pa.attname ORDER BY c.ordinality) AS columns,
                                json_agg(pa2.attname ORDER BY c.ordinality) AS reference_columns
                            FROM
                                unnest(pc.conkey, pc.confkey) WITH ORDINALITY AS c (conkey, confkey, ordinality) 
                                JOIN pg_attribute AS pa ON c.conkey = pa.attnum
                                JOIN pg_attribute AS pa2 ON c.confkey = pa2.attnum
                            WHERE pc.conrelid = pa.attrelid AND pc.confrelid = pa2.attrelid
                        ) AS pa
                    WHERE pc.contype = 'f' AND pn.nspname NOT LIKE 'pg_%'
                ) AS pc2
        """
    )

    (schema_json,) = cur.fetchone()
    return schema_json



import argparse
import logging

from ..log import TRACE
from ..version import __version__


class LogLevel:
    ERROR = "error"
    INFO = "info"
    DEBUG = "debug"
    TRACE = "trace"


def main():
    parser = create_parser()
    args = parser.parse_args()

    setup_logging(args)

    if args.command == "dump":
        from .dump import dump_main

        dump_main(args)
    elif args.command == "restore":
        from .restore import restore_main

        restore_main(args)
    elif args.command == "schema":
        from .schema import schema_main

        schema_main(args)
    elif args.command == "schema-filter":
        from .schema_filter import filter_main

        filter_main(args)
    elif args.command == "transform":
        from .transform import transform_main

        transform_main(args)
    elif args.command == "transform-field":
        from .transform_field import transform_field_main

        transform_field_main(args)


def create_parser():
    parser = argparse.ArgumentParser(
        description="Capture, scrub, and restore subsets of PostgreSQL databases.",
        fromfile_prefix_chars="@",
        formatter_class=ArgumentFormatter,
    )
    update_help(parser)
    parser.add_argument(
        "--log-level",
        choices=[LogLevel.ERROR, LogLevel.INFO, LogLevel.DEBUG, LogLevel.TRACE],
        default=LogLevel.INFO,
        help="Log level (default: %(default)s).",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="Show version and exit.",
    )

    subparsers = parser.add_subparsers(
        dest="command",
        title="subcommands",
        description="Provide one of the subcommands for more specific help.",
    )

    add_dump_command(subparsers)
    add_restore_command(subparsers)
    add_schema_command(subparsers)
    add_schema_filter_command(subparsers)
    add_transform_command(subparsers)
    add_transform_field_command(subparsers)

    return parser


def add_dump_command(subparsers):
    parser = subparsers.add_parser(
        "dump",
        description="Dump data from database.",
        formatter_class=ArgumentFormatter,
    )
    update_help(parser)
    parser.add_argument(
        "--include-schema",
        "--no-include-schema",
        action=NegateAction,
        nargs=0,
        default=False,
        help="Whether to include schema. Only compatible with SQL output.",
    )
    parser.add_argument(
        "-j",
        "--jobs",
        default=1,
        help="Number of workers (default: %(default)d).",
        type=int,
    )
    parser.add_argument(
        "-o",
        "--output",
        default="-",
        help="Path to output slice, or - for stdout (default: %(default)s).",
    )
    parser.add_argument(
        "--output-type", choices=["slice", "sql"], default="slice", help="Output type."
    )
    parser.add_argument(
        "--pepper", help="Pepper to use for transform. Autogenerated if not provided."
    )
    parser.add_argument("--transform", help="Path to transform config, or - for stdin")
    parser.add_argument(
        "-r",
        "--root",
        action="append",
        default=[],
        dest="roots",
        metavar=("TABLE", "CONDITION"),
        nargs=2,
        help="The ID of the root table and SQL condition. May be repeated.",
    )
    parser_required = parser.add_argument_group("required arguments")
    parser_required.add_argument(
        "-s", "--schema", required=True, help="Path to schema, or - for stdin."
    )


def add_restore_command(subparsers):
    parser = subparsers.add_parser(
        "restore", description="Restore data.", formatter_class=ArgumentFormatter
    )
    update_help(parser)
    parser.add_argument(
        "--disable-triggers",
        "--no-disable-triggers",
        action=NegateAction,
        nargs=0,
        default=False,
        help="Disable triggers, including foreign keys.",
    )
    parser.add_argument(
        "-j",
        "--jobs",
        default=1,
        help="Number of workers (default: %(default)d). Using more than one worker requires disabling transactions.",
        type=int,
    )
    parser.add_argument(
        "--include-schema",
        "--no-include-schema",
        default=False,
        help="Include schema (default: %(default)s).",
        action=NegateAction,
        nargs=0,
    )
    parser.add_argument(
        "-i",
        "--input",
        default="-",
        help="Path to slice, or - for stdin (default: %(default)s).",
    )
    parser.add_argument(
        "--transaction",
        "--no-transaction",
        default=True,
        help="Whether to run in a single transaction (default: %(default)s).",
        action=NegateAction,
        nargs=0,
    )


def add_schema_command(subparsers):
    parser = subparsers.add_parser(
        "schema",
        description="Collect schema from database.",
        formatter_class=ArgumentFormatter,
    )
    update_help(parser)
    parser.add_argument(
        "-o",
        "--output",
        default="-",
        help="Path to output schema, or - for stdout (default: %(default)s).",
    )


def add_schema_filter_command(subparsers):
    parser = subparsers.add_parser("schema-filter")
    parser.add_argument("-i", "--input", default="-", help="Input")
    parser.add_argument("-o", "--output", default="-", help="Output")

    subsubparsers = parser.add_subparsers(dest="subcommand", title="subcommands")
    children_parser = subsubparsers.add_parser(
        "children",
        description="Include children, recursively",
        formatter_class=ArgumentFormatter,
    )
    update_help(children_parser)
    children_parser.add_argument(
        "--parents",
        default="none",
        choices=["none", "parents", "intelligent"],
        action=NegateAction,
        nargs=0,
    )
    children_parser.add_argument("table", nargs="*")


def add_transform_command(subparsers):
    parser = subparsers.add_parser(
        "transform", description="Transform slice", formatter_class=ArgumentFormatter
    )
    update_help(parser)
    parser.add_argument("--transform", required=True)


def add_transform_field_command(subparsers):
    parser = subparsers.add_parser(
        "transform-field",
        description="Transform field",
        formatter_class=ArgumentFormatter,
    )
    update_help(parser)
    parser.add_argument("--pepper", help="Pepper.")
    parser.add_argument("--transform", required=True)
    parser.add_argument("field")


def setup_logging(args):
    logging.addLevelName(TRACE, "TRACE")
    logging.basicConfig(format="[%(levelname)-8s] %(message)s")
    if args.log_level == LogLevel.ERROR:
        logging.getLogger().setLevel(logging.ERROR)
    elif args.log_level == LogLevel.INFO:
        logging.getLogger().setLevel(logging.INFO)
    elif args.log_level == LogLevel.DEBUG:
        logging.getLogger().setLevel(logging.DEBUG)
    elif args.log_level == LogLevel.TRACE:
        logging.getLogger().setLevel(TRACE)


class ArgumentFormatter(argparse.RawDescriptionHelpFormatter):
    def __init__(self, prog):
        super().__init__(prog, max_help_position=100, width=100)


class NegateAction(argparse.Action):
    def __call__(self, parser, ns, values, option):
        setattr(ns, self.dest, option[2:4] != "no")


def update_help(parser):
    parser._actions[-1].help = "Show this help message and exit."



import typing
from contextlib import contextmanager

import psycopg2
import psycopg2.sql as sql

Snapshot = str

Tid = str


@contextmanager
def connection(*args, **kwargs):
    conn = psycopg2.connect(*args, **kwargs)
    try:
        yield conn
    finally:
        conn.close()


@contextmanager
def transaction(conn):
    with conn, conn.cursor() as cur:
        yield cur


def export_snapshot(cur) -> Snapshot:
    cur.execute("SELECT pg_export_snapshot()")
    (snapshot,) = cur.fetchone()
    return snapshot


def freeze_transaction(cur, snapshot=None):
    cur.execute("SET TRANSACTION ISOLATION LEVEL REPEATABLE READ READ ONLY")
    if snapshot is not None:
        cur.execute("SET TRANSACTION SNAPSHOT %s", [snapshot])


def defer_constraints(cur, names: typing.List[typing.Tuple[str, str]]):
    query = sql.SQL("SET CONSTRAINTS {} DEFERRED").format(
        sql.SQL(", ").join(sql.Identifier(*name) for name in names)
    )
    cur.execute(query)


def tid_to_int(id: Tid) -> int:
    """
    Translate TID to 48-bit integer, where 32 top significant bits are block number,
    and the other 16 bits are the position within the block
    """
    a, b = map(int, id[1:-1].split(","))
    return a * (2 ** (4 * 8)) + b



import json
import os
import subprocess

from file import temp_file
from pg import connection, transaction
from process import run_process


def test_dump_sql(pg_database, snapshot):
    schema_sql = """
        CREATE TABLE parent (
            id int PRIMARY KEY
        );

        CREATE TABLE child (
            id int PRIMARY KEY,
            parent_id int REFERENCES parent (id)
        );
    """

    with temp_file("schema-") as schema_file, temp_file("output-") as output_file:
        with connection("") as conn, transaction(conn) as cur:
            cur.execute(schema_sql)

            cur.execute(
                """
                    INSERT INTO parent (id)
                    VALUES (1), (2);

                    INSERT INTO child (id, parent_id)
                    VALUES (1, 1), (2, 1), (3, 2);
                """
            )

        with open(schema_file, "w") as f:
            schema_json = {
                "references": [
                    {
                        "columns": ["parent_id"],
                        "id": "public.child.child_parent_id_fkey",
                        "referenceColumns": ["id"],
                        "referenceTable": "public.parent",
                        "table": "public.child",
                    }
                ],
                "tables": [
                    {
                        "columns": ["id"],
                        "id": "public.parent",
                        "name": "parent",
                        "schema": "public",
                    },
                    {
                        "columns": ["id", "parent_id"],
                        "id": "public.child",
                        "name": "child",
                        "schema": "public",
                    },
                ],
            }
            json.dump(schema_json, f)

        run_process(
            [
                "slicedb",
                "dump",
                "--include-schema",
                "--schema",
                schema_file,
                "--root",
                "public.parent",
                "id = 1",
                "--output",
                output_file,
                "--output-type",
                "sql",
            ]
        )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute(
                """
                    DROP TABLE child;

                    DROP TABLE parent;
                """
            )

        with open(output_file) as f:
            print(f.read())

        run_process(
            [
                "psql",
                "-f",
                output_file,
            ],
            env=dict(**os.environ, ON_ERROR_STOP="1"),
        )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("TABLE parent")
            result = cur.fetchall()
            assert result == [(1,)]

            cur.execute("TABLE child")
            result = cur.fetchall()
            assert result == [(1, 1), (2, 1)]



import secrets
import sys

from ..transform import create_transform


def transform_field_main(args):
    transform = create_transform(args.transform)

    if args.pepper is not None:
        pepper = args.pepper.encode("ascii")
    else:
        pepper = secrets.token_bytes(8)

    print(transform.transform(args.field, pepper))



import contextlib

from ..pg import connection
from ..restore import RestoreParams, restore
from .common import open_bytes_read


def restore_main(args):
    params = RestoreParams(include_schema=args.include_schema, parallelism=args.jobs, transaction=args.transaction)

    restore(lambda: connection(""), params, lambda: open_bytes_read(args.input))



import dataclasses
import enum
import typing

import dataclasses_json

from ..json import DataJsonFormat, package_json_format


class DumpReferenceDirection(enum.Enum):
    FORWARD = "forward"
    REVERSE = "reverse"


@dataclasses_json.dataclass_json(
    letter_case=dataclasses_json.LetterCase.CAMEL,
    undefined=dataclasses_json.Undefined.EXCLUDE,
)
@dataclasses.dataclass
class DumpReference:
    columns: typing.List[str]
    id: str
    reference_columns: typing.List[str]
    reference_table: str
    table: str
    directions: typing.List[DumpReferenceDirection] = (
        DumpReferenceDirection.FORWARD,
        DumpReferenceDirection.REVERSE,
    )
    deferrable: bool = False


@dataclasses_json.dataclass_json(
    letter_case=dataclasses_json.LetterCase.CAMEL,
    undefined=dataclasses_json.Undefined.EXCLUDE,
)
@dataclasses.dataclass
class DumpTable:
    columns: typing.List[str]
    id: str
    name: str
    schema: typing.Optional[str]


@dataclasses_json.dataclass_json(
    letter_case=dataclasses_json.LetterCase.CAMEL,
    undefined=dataclasses_json.Undefined.EXCLUDE,
)
@dataclasses.dataclass
class DumpSchema:
    references: typing.List[DumpReference]
    tables: typing.List[DumpTable]


@dataclasses.dataclass
class DumpRoot:
    table: str
    condition: str


DUMP_JSON_FORMAT = package_json_format("slice_db.formats", "dump.json")


DUMP_DATA_JSON_FORMAT = DataJsonFormat(DUMP_JSON_FORMAT, DumpSchema.schema())



import json
import os
import subprocess

from file import temp_file
from pg import connection, transaction
from process import run_process

_SCHEMA_SQL = """
CREATE TABLE directory (
    id int PRIMARY KEY,
    parent_id int REFERENCES directory (id) DEFERRABLE
);
"""

_SCHEMA_JSON = {
    "references": [
        {
            "columns": ["parent_id"],
            "id": "public.directory.directory_parent_id_fkey",
            "referenceColumns": ["id"],
            "referenceTable": "public.directory",
            "table": "public.directory",
        }
    ],
    "tables": [
        {
            "columns": ["id", "parent_id"],
            "id": "public.directory",
            "name": "directory",
            "schema": "public",
        },
    ],
}


def test_defer(pg_database, snapshot):
    with temp_file("schema-") as schema_file, temp_file("output-") as output_file:
        with connection("") as conn, transaction(conn) as cur:
            cur.execute(_SCHEMA_SQL)

            cur.execute(
                """
                INSERT INTO directory (id, parent_id)
                VALUES (1, NULL), (2, 1), (3, 1);
                """
            )

        with open(schema_file, "w") as f:
            json.dump(_SCHEMA_JSON, f)

        run_process(
            [
                "slicedb",
                "dump",
                "--schema",
                schema_file,
                "--root",
                "public.directory",
                "id = 1",
                "--output",
                output_file,
            ]
        )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute(
                """
                SET CONSTRAINTS ALL DEFERRED;

                DELETE FROM directory;
                """
            )

        run_process(
            [
                "slicedb",
                "restore",
                "--input",
                output_file,
            ]
        )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("SELECT * FROM directory ORDER BY id")
            result = cur.fetchall()
            assert result == [(1, None), (2, 1), (3, 1)]


def test_defer_cycle(pg_database, snapshot):
    with temp_file("schema-") as schema_file, temp_file("output-") as output_file:
        with connection("") as conn, transaction(conn) as cur:
            cur.execute(_SCHEMA_SQL)

            cur.execute(
                """
                INSERT INTO directory (id, parent_id)
                VALUES (1, 2), (2, 3), (3, 1);
                """
            )

        with open(schema_file, "w") as f:
            json.dump(_SCHEMA_JSON, f)

        run_process(
            [
                "slicedb",
                "dump",
                "--schema",
                schema_file,
                "--root",
                "public.directory",
                "id = 1",
                "--output",
                output_file,
            ]
        )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute(
                """
                SET CONSTRAINTS ALL DEFERRED;

                DELETE FROM directory;
                """
            )

        run_process(
            [
                "slicedb",
                "restore",
                "--input",
                output_file,
            ]
        )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("SELECT * FROM directory ORDER BY id")
            result = cur.fetchall()
            assert result == [(1, 2), (2, 3), (3, 1)]



from ..dump import DumpIo, DumpParams, OutputType, dump
from ..formats.dump import DumpRoot
from ..pg import connection
from .common import open_bytes_write, open_str_read


def dump_main(args):
    roots = [
        DumpRoot(condition=condition, table=table) for table, condition in args.roots
    ]

    if args.output_type == "slice":
        output_type = OutputType.SLICE
    elif args.output_type == "sql":
        output_type = OutputType.SQL

    io = DumpIo(
        conn=lambda: connection(""),
        output=lambda: open_bytes_write(args.output),
        schema_file=lambda: open_str_read(args.schema),
    )
    params = DumpParams(
        include_schema=args.include_schema,
        parallelism=args.jobs,
        output_type=output_type,
    )

    dump(roots, io, params)



from ..formats.dump import DUMP_JSON_FORMAT
from ..pg import connection, transaction
from ..schema import query_schema
from .common import open_str_write


def schema_main(args):
    with connection("") as conn, transaction(conn) as cur:
        schema_json = query_schema(cur)

    DUMP_JSON_FORMAT.dump(lambda: open_str_write(args.output), schema_json, pretty=True)



import contextlib
import dataclasses
import itertools
import json
import logging
import time
import typing
import zipfile

import psycopg2.sql as sql

from .collection.dict import groups
from .concurrent.graph import GraphRunner
from .formats.dump import DumpSchema
from .formats.manifest import (
    MANIFEST_DATA_JSON_FORMAT,
    Manifest,
    ManifestTable,
    ManifestTableSegment,
)
from .log import TRACE
from .pg import defer_constraints, transaction
from .resource import NoArgs
from .slice import SliceReader


@dataclasses.dataclass
class RestoreParams:
    include_schema: bool
    parallelism: int
    transaction: bool


def restore(conn_fn, params, file_fn):
    if not params.include_schema and params.parallelism > 1 and params.transaction:
        raise Exception("A single transaction must be disabled for parallelism > 1")

    with file_fn() as file, SliceReader(file) as reader:
        manifest = MANIFEST_DATA_JSON_FORMAT.load(reader.open_manifest)
        manifest_tables = {table.id: table for table in manifest.tables}

        items = {id: RestoreItem(table=table) for id, table in manifest_tables.items()}
        restore = Restore(reader)

        with conn_fn() as conn, transaction(conn) as cur:
            if params.transaction:

                def cur_factory():
                    return contextlib.nullcontext(contextlib.nullcontext(cur))

            else:

                @contextlib.contextmanager
                def cur_factory():
                    with conn_fn() as conn:
                        yield NoArgs(lambda: transaction(conn))

            runner = GraphRunner(params.parallelism, restore.process, cur_factory)

            if params.include_schema:
                constraints = []
            else:
                constraints = get_constaints(cur, list(manifest_tables.values()))

            deferrable_constaints = [
                [constraint.schema, constraint.name]
                for constraint in constraints
                if constraint.deferrable
            ]

            if deferrable_constaints:
                logging.info("Deferring %d constraints", len(deferrable_constaints))
                defer_constraints(cur, deferrable_constaints)

            deps = groups(
                (constraint for constraint in constraints if not constraint.deferrable),
                lambda constraint: constraint.table,
            )

            if params.include_schema:
                with cur_factory() as conn, conn as cur:
                    with reader.open_schema("pre-data") as f:
                        schema_sql = f.read()
                    cur.execute(schema_sql)

            runner.run(
                list(items.values()),
                lambda item: [
                    items[foreign_key.reference_table]
                    for foreign_key in deps[item.table.id]
                ],
            )

            if params.include_schema:
                with cur_factory() as conn, conn as cur:
                    with reader.open_schema("post-data") as f:
                        schema_sql = f.read()
                    cur.execute(schema_sql)


@dataclasses.dataclass
class RestoreItem:
    table: ManifestTable

    def __hash__(self):
        return id(self)


class Restore:
    def __init__(self, slice_reader: SliceReader):
        self._slice_reader = slice_reader

    def process(self, item: RestoreItem, transaction):
        with transaction as cur:
            for i, segment in enumerate(item.table.segments):
                with self._slice_reader.open_segment(
                    item.table.id,
                    i,
                ) as file:
                    update_data(cur, item.table, i, segment, file)


_BUFFER_SIZE = 1024 * 32


def update_data(
    cur, table: ManifestTable, index: int, segment: ManifestTableSegment, in_
):
    logging.log(TRACE, f"Restoring %s rows into table %s", segment.row_count, table.id)
    start = time.perf_counter()
    cur.copy_from(
        in_,
        sql.Identifier(table.schema, table.name).as_string(cur),
        columns=[sql.Identifier(column).as_string(cur) for column in table.columns],
        size=_BUFFER_SIZE,
    )
    end = time.perf_counter()
    logging.debug(
        f"Restored %s rows in table %s (%.3fs)",
        segment.row_count,
        table.id,
        end - start,
    )


@dataclasses.dataclass
class ForeignKey:
    deferrable: bool
    """Deferrable"""
    name: str
    """Name"""
    schema: str
    """Schema"""
    table: str
    """Table ID"""
    reference_table: str
    """Referenced table ID"""


def get_constaints(
    cur, manifest_tables: typing.List[ManifestTable]
) -> typing.List[ForeignKey]:
    """
    Query PostgreSQL for constraints between tables
    """
    cur.execute(
        """
            WITH
                "table" AS (
                    SELECT *
                    FROM unnest(%s::text[], %s::text[], %s::text[]) AS t (id, schema, name)
                )
            SELECT
                pn.nspname,
                pc.conname,
                a.id,
                b.id,
                pc.condeferrable
            FROM
                pg_constraint AS pc
                JOIN pg_class AS pc2 ON pc.conrelid = pc2.oid
                JOIN pg_namespace AS pn ON pc2.relnamespace = pn.oid
                JOIN "table" AS a ON (pn.nspname, pc2.relname) = (a.schema, a.name)
                JOIN pg_class AS pc3 ON pc.confrelid = pc3.oid
                JOIN pg_namespace AS pn2 ON pc3.relnamespace = pn2.oid
                JOIN "table" AS b ON (pn2.nspname, pc3.relname) = (b.schema, b.name)
            WHERE pc.contype = 'f'
        """,
        [
            [table.id for table in manifest_tables],
            [table.schema for table in manifest_tables],
            [table.name for table in manifest_tables],
        ],
    )

    foreign_keys = []
    for schema, name, table, reference_table, deferrable in cur.fetchall():
        foreign_keys.append(
            ForeignKey(
                deferrable=deferrable,
                name=name,
                reference_table=reference_table,
                schema=schema,
                table=table,
            )
        )

    return foreign_keys



import dataclasses
import typing

import dataclasses_json

from ..json import DataJsonFormat, package_json_format


@dataclasses_json.dataclass_json(letter_case=dataclasses_json.LetterCase.CAMEL)
@dataclasses.dataclass()
class ManifestTableSegment:
    row_count: int


@dataclasses_json.dataclass_json(letter_case=dataclasses_json.LetterCase.CAMEL)
@dataclasses.dataclass()
class ManifestTable:
    columns: typing.List[str]
    """Columns"""
    id: str
    """ID"""
    name: str
    """Name"""
    schema: str
    """Schema"""
    segments: typing.List[ManifestTableSegment]
    """Segments"""


@dataclasses_json.dataclass_json(letter_case=dataclasses_json.LetterCase.CAMEL)
@dataclasses.dataclass()
class Manifest:
    tables: typing.List[ManifestTable]


MANIFEST_JSON_FORMAT = package_json_format("slice_db.formats", "manifest.json")


MANIFEST_DATA_JSON_FORMAT = DataJsonFormat(MANIFEST_JSON_FORMAT, Manifest.schema())



#!/usr/bin/env python3
import os

import setuptools

version = {}
with open("slice_db/version.py", "r") as f:
    exec(f.read(), version)

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    author="Rivet Health",
    author_email="ops@rivethealth.com",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    description="Translate between MLLP and HTTP",
    entry_points={
        "console_scripts": [
            "slicedb=slice_db.cli.main:main",
        ]
    },
    extras_require={"dev": ["black", "pytest-env", "isort", "pytest", "snapshottest"]},
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["dataclasses_json", "numpy"],
    name="slice-db",
    packages=setuptools.find_packages(),
    project_urls={
        "Issues": "https://github.com/rivethealth/slice-db/issues",
    },
    python_requires=">3.7.0",
    url="https://github.com/rivethealth/slice-db",
    version=version["__version__"],
)



from __future__ import annotations

import collections
import contextlib
import dataclasses
import enum
import logging
import shutil
import subprocess
import tempfile
import threading
import time
import typing

import numpy
import psycopg2.sql as sql

from .collection.set import IntSet
from .concurrent.work import Worker, WorkerRunner
from .formats.dump import (
    DUMP_DATA_JSON_FORMAT,
    DumpReferenceDirection,
    DumpRoot,
    DumpSchema,
)
from .formats.manifest import (
    MANIFEST_DATA_JSON_FORMAT,
    Manifest,
    ManifestTable,
    ManifestTableSegment,
)
from .log import TRACE
from .pg import Tid, export_snapshot, freeze_transaction, tid_to_int, transaction
from .resource import ResourceFactory
from .slice import SliceWriter
from .sql import SqlWriter


class OutputType(enum.Enum):
    SQL = enum.auto()
    SLICE = enum.auto()


@dataclasses.dataclass
class DumpIo:
    conn: ResourceFactory
    schema_file: ResourceFactory[typing.TextIO]
    output: ResourceFactory[typing.BinaryIO]


@dataclasses.dataclass
class DumpParams:
    include_schema: bool
    parallelism: int
    output_type: OutputType


def dump(
    root_configs: typing.List[DumpRoot],
    io: DumpIo,
    params: DumpParams,
):
    """
    Dump
    """
    dump_config = DUMP_DATA_JSON_FORMAT.load(io.schema_file)
    schema = Schema(dump_config)
    roots = []
    for root_config in root_configs:
        try:
            table = schema.get_table(root_config.table)
        except KeyError:
            raise Exception(f"Root table {root_config.table} does not exist")
        roots.append(Root(table=table, condition=sql.SQL(root_config.condition)))

    with io.output() as file, io.conn() as conn, contextlib.ExitStack() as stack:
        if params.output_type == OutputType.SLICE:
            slice_writer = stack.enter_context(SliceWriter(file))
            output = _SliceOutput(slice_writer)
        elif params.output_type == OutputType.SQL:
            sql_writer = SqlWriter(conn, file)
            if params.include_schema:
                # must add schema before others
                with sql_writer.open_predata() as f:
                    _pg_dump_section("pre-data", f)
            output = _SqlOutput(sql_writer)

        result = _DiscoveryResult()

        with transaction(conn) as cur:
            if params.parallelism == 1:
                freeze_transaction(cur)

                def pg_manager():
                    return contextlib.nullcontext(cur)

            else:
                snapshot = export_snapshot(cur)
                logging.info("Running at snapshot %s", snapshot)

                @contextlib.contextmanager
                def pg_manager():
                    with io.conn() as conn, transaction(conn) as cur:
                        freeze_transaction(cur, snapshot=snapshot)
                        yield cur

            _dump_rows(
                roots,
                params.include_schema and params.output_type != OutputType.SQL,
                params.parallelism,
                pg_manager,
                result,
                output,
            )

        if params.output_type == OutputType.SLICE:
            manifest = Manifest(tables=result.table_manifests())
            MANIFEST_DATA_JSON_FORMAT.dump(slice_writer.open_manifest, manifest)
        elif params.output_type == OutputType.SQL:
            if params.include_schema:
                with sql_writer.open_postdata() as f:
                    _pg_dump_section("post-data", f)


def _dump_rows(
    roots: typing.List[Root],
    include_schema: bool,
    parallelism: int,
    cur_resource: ResourceFactory,
    result,
    output: typing.Union[_SliceOutput, _SqlOutput],
):
    """
    Dump rows
    """

    if include_schema:
        logging.info("Dumping schema and rows")
    else:
        logging.info("Dumping rows")
    start = time.perf_counter()
    worker = _Dump(result, output)
    runner = WorkerRunner(parallelism, worker.process_item, cur_resource)
    items = []
    if include_schema:
        items.append(_Dump.SchemaItem(section="pre-data"))
        items.append(_Dump.SchemaItem(section="post-data"))
    items += [
        _Dump.RootItem(table=root.table, condition=root.condition) for root in roots
    ]
    runner.run(items)
    end = time.perf_counter()
    if include_schema:
        logging.info("Dumped schema and %d total rows (%.3fs)", result.row_count, end - start)
    else:
        logging.info("Dumped %d total rows (%.3fs)", result.row_count, end - start)


class _SliceOutput:
    """
    Thread-safe slice output
    """

    def __init__(self, writer: SliceWriter):
        self._lock = threading.Lock()
        self._writer = writer

    @contextlib.contextmanager
    def open_schema(self, section: str):
        """
        Open schema for writing
        """
        with self._lock, self._writer.open_schema(section) as f:
            yield f

    @contextlib.contextmanager
    def open_segment(self, segment: TableSegment):
        """
        Open segment for writing
        """
        with self._lock, self._writer.open_segment(
            segment.table.id, segment.index
        ) as f:
            yield f


class _SqlOutput:
    """
    Thread-safe SQL output
    """

    def __init__(self, writer: SqlWriter):
        self._lock = threading.Lock()
        self._writer = writer

    def open_schema(self, section: str):
        raise Exception("Not supported")

    @contextlib.contextmanager
    def open_segment(self, segment: TableSegment):
        with self._lock, self._writer.open_data(
            segment.table.id,
            segment.index,
            segment.table.schema,
            segment.table.name,
            segment.table.columns,
        ) as f:
            yield f


def _pg_dump_section(section: str, out: typing.BinaryIO) -> str:
    logging.log(TRACE, "Dumping %s schema", section)
    start = time.perf_counter()
    subprocess.check_call(
        ["pg_dump", "-B", "--no-acl", "--section", section],
        stdin=subprocess.DEVNULL,
        stdout=out,
    )
    end = time.perf_counter()
    logging.debug("Dumped %s schema (%.3fs)", section, end - start)


class _DiscoveryResult:
    """
    Discovered IDs
    """

    _row_ids: typing.DefaultDict[str, IntSet]
    _table_manifests: typing.Dict[str, ManifestTable]

    def __init__(self):
        self._id_count = 0
        self._row_ids = collections.defaultdict(lambda: IntSet(numpy.int64))
        self._table_manifests = {}
        self._lock = threading.Lock()

    def add(
        self, table: Table, row_ids: typing.List[Tid]
    ) -> typing.Optional[TableSegment]:
        """
        Add IDs and return list of newly added segment
        """
        with self._lock:
            existing_ids = self._row_ids[table.id]
            ints = [tid_to_int(id) for id in row_ids]
            contains = existing_ids.contains(ints)
            new_ids = [id for id, c in zip(row_ids, contains) if not c]

            if not new_ids:
                return

            existing_ids.add([int for int, c in zip(ints, contains) if not c])
            self._id_count += len(new_ids)

            if table.id not in self._table_manifests:
                self._table_manifests[table.id] = ManifestTable(
                    columns=table.columns,
                    id=table.id,
                    name=table.name,
                    schema=table.schema,
                    segments=[],
                )
            table_manifest = self._table_manifests[table.id]

            segment = TableSegment(
                table=table, row_ids=new_ids, index=len(table_manifest.segments)
            )
            table_manifest.segments.append(ManifestTableSegment(row_count=len(new_ids)))

        return segment

    @property
    def row_count(self):
        """
        Total rows
        """
        return self._id_count

    def table_manifests(self):
        """
        Iterable of ManifestTables
        """
        return self._table_manifests.values()


class _Dump:
    @dataclasses.dataclass
    class RootItem:
        table: Table
        """Table"""
        condition: sql.SQL
        """Condition"""

    @dataclasses.dataclass
    class ReferenceItem:
        direction: DumpReferenceDirection
        """Direction"""
        reference: Reference
        """Reference"""
        segment: TableSegment
        """Source segment"""

    @dataclasses.dataclass
    class SchemaItem:
        section: str
        """Section"""

    def __init__(self, result: _DiscoveryResult, output: _SliceOutput):
        self._result = result
        self._output = output

    def process_item(
        self, item: typing.Union[_Dump.RootItem, _Dump.ReferenceItem], cur
    ):
        """
        Process item
        """
        if isinstance(item, _Dump.RootItem):
            segment = _discover_table_condition(
                cur, item.table, item.condition, self._result
            )
            if segment is None:
                return
            to_table = item.table

            yield from self._table_items(segment)
        elif isinstance(item, _Dump.ReferenceItem):
            segment = _discover_reference(
                cur, item.reference, item.direction, item.segment, self._result
            )
            if segment is None:
                return

            if item.direction == DumpReferenceDirection.FORWARD:
                to_table = item.reference.reference_table
            elif item.direction == DumpReferenceDirection.REVERSE:
                to_table = item.reference.table

            yield from self._table_items(segment, reference_item=item)
        elif isinstance(item, _Dump.SchemaItem):
            with tempfile.TemporaryFile() as tmp:
                _pg_dump_section(item.section, tmp)
                tmp.seek(0)
                with self._output.open_schema(item.section) as f:
                    shutil.copyfileobj(tmp, f)
                return

        with tempfile.TemporaryFile() as tmp:
            _dump_data(cur, to_table, segment.row_ids, tmp)
            tmp.seek(0)
            with self._output.open_segment(segment) as f:
                shutil.copyfileobj(tmp, f)

    def _table_items(
        self,
        segment: TableSegment,
        reference_item: _Dump.ReferenceItem = None,
    ):
        """
        Create items for table
        """
        for reference in segment.table.references:
            if DumpReferenceDirection.FORWARD not in reference.directions:
                continue
            if (
                reference_item is not None
                and reference is reference_item.reference
                and reference_item.direction == DumpReferenceDirection.REVERSE
            ):
                continue
            yield _Dump.ReferenceItem(
                segment=segment,
                reference=reference,
                direction=DumpReferenceDirection.FORWARD,
            )
        for reference in segment.table.reverse_references:
            if DumpReferenceDirection.REVERSE not in reference.directions:
                continue
            if (
                reference_item is not None
                and reference is reference_item.reference
                and reference_item.direction == DumpReferenceDirection.FORWARD
            ):
                continue
            yield _Dump.ReferenceItem(
                segment=segment,
                reference=reference,
                direction=DumpReferenceDirection.REVERSE,
            )


@dataclasses.dataclass
class Root:
    """Root"""

    table: Table
    """Table"""
    condition: sql.SQL
    """Condition"""


@dataclasses.dataclass
class Table:
    """Table"""

    id: str
    """ID"""
    name: str
    """Name"""
    schema: str
    """Schema"""
    columns: typing.List[str]
    """Columns"""
    references: typing.List[Reference]
    """References to parent tables"""
    reverse_references: typing.List[Reference]
    """References to child tables"""


@dataclasses.dataclass
class Reference:
    """Reference"""

    directions: typing.List[DumpReferenceDirection]
    """Directions"""
    id: str
    """ID"""
    table: Table
    """Table"""
    columns: typing.List[str]
    """Columns"""
    reference_table: Table
    """Reference columns"""
    reference_columns: typing.List[str]


@dataclasses.dataclass
class TableSegment:
    index: int
    row_ids: typing.List[Tid]
    table: Table


class Schema:
    """
    Graph model of schema
    """

    def __init__(self, schema: DumpSchema):
        self._tables = {}
        for table_config in schema.tables:
            table = Table(
                columns=table_config.columns,
                references=[],
                id=table_config.id,
                name=table_config.name,
                reverse_references=[],
                schema=table_config.schema,
            )
            if table.id in self._tables:
                raise Exception(f"Multiple definitions for table {table.id}")
            self._tables[table.id] = table

        self._references = {}
        for reference_config in schema.references:
            try:
                table = self._tables[reference_config.table]
            except KeyError:
                raise Exception(
                    f"No table {reference_config.table}, needed by reference {reference_config.id}"
                )
            try:
                reference_table = self._tables[reference_config.reference_table]
            except KeyError:
                raise Exception(
                    f"No table {reference_config.reference_table}, needed by reference {reference_config.id}"
                )

            reference = Reference(
                directions=reference_config.directions,
                id=reference_config.id,
                table=table,
                columns=reference_config.columns,
                reference_table=reference_table,
                reference_columns=reference_config.reference_columns,
            )
            if reference.id in self._references:
                raise Exception(f"Multiple definitions for reference {reference.id}")
            self._references[reference.id] = reference
            table.references.append(reference)
            reference_table.reverse_references.append(reference)

    def get_table(self, id) -> Table:
        """
        Get table by ID
        """
        return self._tables[id]

    def tables(self):
        """
        Iterable of tables
        """
        return self._tables.values()


def _dump_data(cur, table: Table, ids: typing.List[Tid], out):
    """
    Dump data
    """

    logging.log(TRACE, f"Dumping %s rows from table %s", len(ids), table.id)
    start = time.perf_counter()
    query = sql.SQL(
        """
            COPY (
                SELECT {}
                FROM {}
                WHERE ctid = ANY({}::tid[])
            )
            TO STDOUT
        """
    ).format(
        sql.SQL(", ").join([sql.Identifier(column) for column in table.columns]),
        sql.Identifier(table.schema, table.name),
        sql.Literal(ids),
    )
    cur.copy_expert(query, out, size=1024 * 32)
    end = time.perf_counter()
    logging.debug(
        f"Dumped %s rows from table %s (%.3fs)", len(ids), table.id, end - start
    )


def _discover_table_condition(
    cur, table: Table, condition: sql.SQL, result: _DiscoveryResult
) -> typing.List[Tid]:
    """
    Discover, using root
    """
    logging.log(TRACE, f"Finding rows from table %s", table.id)
    start = time.perf_counter()
    query = sql.SQL(
        """
            SELECT ctid
            FROM {}
            WHERE {}
        """
    ).format(
        sql.Identifier(table.schema, table.name),
        condition,
    )
    cur.execute(query)
    found_ids = [id_ for id_, in cur.fetchall()]
    segment = result.add(table, found_ids) if found_ids else None
    end = time.perf_counter()
    if segment is None:
        logging.debug(
            f"Found %s rows (no new) in table %s (%.3fs)",
            len(found_ids),
            table.id,
            end - start,
        )
    else:
        logging.debug(
            f"Found %s rows (%s new) as %s/%s (%.3fs)",
            len(found_ids),
            len(segment.row_ids),
            segment.table.id,
            segment.index,
            end - start,
        )
    end = time.perf_counter()
    return segment


def _discover_reference(
    cur,
    reference: Reference,
    direction: DumpReferenceDirection,
    segment: TableSegment,
    result,
) -> typing.List[Tid]:
    """
    Discover, using reference
    """
    if direction == DumpReferenceDirection.FORWARD:
        from_columns = reference.columns
        from_table = reference.table
        to_columns = reference.reference_columns
        to_table = reference.reference_table
    elif direction == DumpReferenceDirection.REVERSE:
        from_columns = reference.reference_columns
        from_table = reference.reference_table
        to_columns = reference.columns
        to_table = reference.table

    logging.log(
        TRACE,
        f"Finding rows from table %s using %s/%s via %s",
        to_table.id,
        segment.table.id,
        segment.index,
        reference.id,
    )
    start = time.perf_counter()
    query = sql.SQL(
        """
            SELECT DISTINCT b.ctid
            FROM {} AS a
                JOIN {} AS b ON ({}) = ({})
            WHERE a.ctid = ANY(%s::tid[])
        """
    ).format(
        sql.Identifier(from_table.schema, from_table.name),
        sql.Identifier(to_table.schema, to_table.name),
        sql.SQL(", ").join([sql.Identifier("a", name) for name in from_columns]),
        sql.SQL(", ").join([sql.Identifier("b", name) for name in to_columns]),
    )
    cur.execute(query, [segment.row_ids])
    found_ids = [id_ for id_, in cur.fetchall()]

    new_segment = result.add(to_table, found_ids) if found_ids else None
    end = time.perf_counter()
    if new_segment is None:
        logging.debug(
            f"Found %s rows (no new) in table %s using %s/%s via %s (%.3fs)",
            len(found_ids),
            to_table.id,
            segment.table.id,
            segment.index,
            reference.id,
            end - start,
        )
    else:
        logging.debug(
            f"Found %s rows (%s new) as %s/%s using %s/%s via %s (%.3fs)",
            len(found_ids),
            len(new_segment.row_ids),
            new_segment.table.id,
            new_segment.index,
            segment.table.id,
            segment.index,
            reference.id,
            end - start,
        )

    return new_segment



import json
import typing

import dataclasses_json.mm
import jsonschema

from ..resource import ResourceFactory

try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources


T = typing.TypeVar("T")


class JsonSchema:
    def __init__(self, schema):
        self._schema = schema

    def validate(self, instance):
        jsonschema.validate(instance, self._schema)


class JsonFormat:
    """
    Parses and serializes JSON, validating via JSONSchema.
    """

    def __init__(self, json_schema: JsonSchema):
        self._json_schema = json_schema

    def load(self, file_fn: ResourceFactory[typing.TextIO]):
        with file_fn() as file:
            instance = json.load(file)
        self._json_schema.validate(instance)
        return instance

    def dump(self, file_fn: ResourceFactory[typing.TextIO], instance, pretty=False):
        self._json_schema.validate(instance)
        with file_fn() as file:
            json.dump(instance, file, sort_keys=True, indent=2 if pretty else None)


class DataJsonFormat(typing.Generic[T]):
    """
    Coverts to dataclasses, while also validating via JSONSchema
    """

    def __init__(self, format: JsonFormat, schema: dataclasses_json.mm.SchemaF):
        self._format = format
        self._dataclass_schema = schema

    def load(self, file_fn: ResourceFactory[typing.TextIO]) -> T:
        instance = self._format.load(file_fn)
        return self._dataclass_schema.load(instance)

    def dump(self, file_fn: ResourceFactory[typing.TextIO], instance: T):
        instance = self._dataclass_schema.dump(instance)
        self._format.dump(file_fn, instance)


def package_json_format(package: str, name: str):
    with pkg_resources.open_text(package, name) as f:
        json_schema_data = json.load(f)
    json_schema = JsonSchema(json_schema_data)
    return JsonFormat(json_schema)



import typing

T = typing.TypeVar("T")

ResourceFactory = typing.Callable[[], typing.ContextManager[T]]


class NoArgs:
    def __init__(self, fn):
        self._fn = fn
        self._resource = None

    def __enter__(self, *args, **kwargs):
        self._resource = self._fn()
        return self._resource.__enter__(*args, **kwargs)

    def __exit__(self, *args, **kwargs):
        self._resource.__exit__(*args, **kwargs)
        self._resource = None



from __future__ import annotations

import typing

T = typing.TypeVar("T")

DepFn = typing.Callable[[T], typing.List[T]]


class CycleError(Exception):
    def __init__(self, nodes):
        super().__init__(CycleError.str(nodes))
        self.nodes = nodes

    @staticmethod
    def str(nodes: typing.List[T]) -> str:
        nodes = nodes + [nodes[0]]
        return " -> ".join(str(node) for node in nodes)


def check_cycle(nodes: typing.List[T], deps_fn: DepFn[T]) -> None:
    """
    Raise CycleError if cycle exists
    """
    visited = set()
    stack = []
    stack_set = set()

    def f(nodes: typing.List[T]):
        for node in nodes:
            if node in visited:
                return

            if node in stack_set:
                i = stack.index(node)
                raise CycleError(stack[i:])

            stack.append(node)
            stack_set.add(node)

            f(deps_fn(node))

            stack.pop()
            stack_set.remove(node)

            visited.add(node)

    f(nodes)
